@@ -21,7 +21,11 @@
     :license: GNU AGPLv3+ or BSD
 
 """
+from __future__ import division
+
+from collections import defaultdict
 import datetime
+import math
 
 from dateutil.parser import parse as parse_datetime
 from flask import abort
@@ -36,11 +40,14 @@ from sqlalchemy.orm import class_mapper
 from sqlalchemy.orm import ColumnProperty
 from sqlalchemy.orm import object_mapper
 from sqlalchemy.orm import RelationshipProperty
+from sqlalchemy.orm.dynamic import AppenderMixin
 from sqlalchemy.orm.exc import MultipleResultsFound
 from sqlalchemy.orm.exc import NoResultFound
 from sqlalchemy.orm.properties import RelationshipProperty as RelProperty
+from sqlalchemy.orm.query import Query
 from sqlalchemy.sql import func
 
+from .helpers import partition
 from .helpers import unicode_keys_to_strings
 from .search import create_query
 from .search import search
@@ -92,15 +99,13 @@ def _get_or_create(session, model, **kwargs):
     :func:`sqlalchemy.orm.query.Query.filter_by` function.
 
     """
-    # TODO document that this uses the .first() function
     instance = session.query(model).filter_by(**kwargs).first()
     if instance:
         return instance, False
-    else:
-        instance = model(**kwargs)
-        session.add(instance)
-        session.commit()
-        return instance, True
+    instance = model(**kwargs)
+    session.add(instance)
+    session.commit()
+    return instance, True
 
 
 def _get_columns(model):
@@ -144,28 +149,50 @@ def _primary_key_name(model_or_instance):
 
 # This code was adapted from :meth:`elixir.entity.Entity.to_dict` and
 # http://stackoverflow.com/q/1958219/108197.
-#
-# TODO should we have an `include` argument also?
-def _to_dict(instance, deep=None, exclude=None):
+def _to_dict(instance, deep=None, exclude=None, include=None,
+             exclude_relations=None, include_relations=None):
     """Returns a dictionary representing the fields of the specified `instance`
     of a SQLAlchemy model.
 
     `deep` is a dictionary containing a mapping from a relation name (for a
     relation of `instance`) to either a list or a dictionary. This is a
     recursive structure which represents the `deep` argument when calling
-    `_to_dict` on related instances. When an empty list is encountered,
-    `_to_dict` returns a list of the string representations of the related
-    instances.
+    :func:`!_to_dict` on related instances. When an empty list is encountered,
+    :func:`!_to_dict` returns a list of the string representations of the
+    related instances.
 
-    `exclude` specifies the columns which will *not* be present in the returned
-    dictionary representation of the object.
+    If either `include` or `exclude` is not ``None``, exactly one of them must
+    be specified. If both are not ``None``, then this function will raise a
+    :exc:`ValueError`. `exclude` must be a list of strings specifying the
+    columns which will *not* be present in the returned dictionary
+    representation of the object (in other words, it is a
+    blacklist). Similarly, `include` specifies the only columns which will be
+    present in the returned dictionary (in other words, it is a whitelist).
+
+    .. note::
+
+       If `include` is an iterable of length zero (like the empty tuple or the
+       empty list), then the returned dictionary will be empty. If `include` is
+       ``None``, then the returned dictionary will include all columns not
+       excluded by `exclude`.
+
+    `include_relations` is a dictionary mapping strings representing relation
+    fields on the specified `instance` to a list of strings representing the
+    names of fields on the related model which should be included in the
+    returned dictionary; `exclude_relations` is similar.
 
     """
-    deep = deep or {}
-    exclude = exclude or ()
+    if (exclude is not None or exclude_relations is not None) and \
+            (include is not None or include_relations is not None):
+        raise ValueError('Cannot specify both include and exclude.')
     # create the dictionary mapping column name to value
     columns = (p.key for p in object_mapper(instance).iterate_properties
                if isinstance(p, ColumnProperty))
+    # filter the columns based on exclude and include values
+    if exclude is not None:
+        columns = (c for c in columns if c not in exclude)
+    elif include is not None:
+        columns = (c for c in columns if c in include)
     result = dict((col, getattr(instance, col)) for col in columns)
     # Convert datetime and date objects to ISO 8601 format.
     #
@@ -174,43 +201,102 @@ def _to_dict(instance, deep=None, exclude=None):
         if isinstance(value, datetime.date):
             result[key] = value.isoformat()
     # recursively call _to_dict on each of the `deep` relations
+    deep = deep or {}
     for relation, rdeep in deep.iteritems():
-        # exclude foreign keys of the related object for the recursive call
-        relationproperty = object_mapper(instance).get_property(relation)
-        newexclude = (key.name for key in relationproperty.remote_side)
-        # get the related value so we can see if it is None or a list
+        # Get the related value so we can see if it is None, a list, a query
+        # (as specified by a dynamic relationship loader), or an actual
+        # instance of a model.
         relatedvalue = getattr(instance, relation)
+        # HACK: In case the relatedvalue is a dynamically loaded
+        # relationship, we need to resolve the query into a concrete
+        # list of objects; see issue #89. We should also check to see
+        # if relatedvalue is a many-to-one relationship, in order to
+        # call relatedvalue.one() or something, but I don't know how
+        # to do that.
+        if isinstance(relatedvalue, (AppenderMixin, Query)):
+            relatedvalue = relatedvalue.all()
         if relatedvalue is None:
             result[relation] = None
-        elif isinstance(relatedvalue, list):
-            result[relation] = [_to_dict(inst, rdeep, newexclude)
-                                for inst in relatedvalue]
         else:
-            result[relation] = _to_dict(relatedvalue, rdeep, newexclude)
+            # Determine the included and excluded fields for the related model.
+            newexclude = None
+            newinclude = None
+            if exclude_relations is not None and relation in exclude_relations:
+                newexclude = exclude_relations[relation]
+            elif include_relations is not None and \
+                    relation in include_relations:
+                newinclude = include_relations[relation]
+            if isinstance(relatedvalue, list):
+                result[relation] = [_to_dict(inst, rdeep, exclude=newexclude,
+                                             include=newinclude)
+                                    for inst in relatedvalue]
+            else:
+                result[relation] = _to_dict(relatedvalue, rdeep,
+                                            exclude=newexclude,
+                                            include=newinclude)
     return result
 
 
-def _include_keys(dictionary, keys):
-    """Returns a new dictionary containing only the mappings from `dictionary`
-    with keys as specified in `keys`.
+def _parse_includes(column_names):
+    """Returns a pair, consisting of a list of column names to include on the
+    left and a dictionary mapping relation name to a list containing the names
+    of fields on the related model which should be included.
+
+    `column_names` is either ``None`` or a list of strings. If it is ``None``,
+    the returned pair will be ``(None, None)``.
+
+    If the name of a relation appears as a key in the dictionary, then it will
+    not appear in the list.
 
     """
-    return dict((k, v) for k, v in dictionary.items() if k in keys)
+    if column_names is None:
+        return None, None
+    dotted_names, columns = partition(column_names, lambda name: '.' in name)
+    # Create a dictionary mapping relation names to fields on the related
+    # model.
+    relations = defaultdict(list)
+    for name in dotted_names:
+        relation, field = name.split('.', 1)
+        # Only add the relation if it's column has been specified.
+        if relation in columns:
+            relations[relation].append(field)
+    # Included relations need only be in the relations dictionary, not the
+    # columns list.
+    for relation in relations:
+        if relation in columns:
+            columns.remove(relation)
+    return columns, relations
 
 
-def _to_dict_include(instance, deep=None, exclude=None, include=None):
-    """Returns the dictionary representation of `instance` as returned by
-    :func:`_to_dict`, but only with the fields specified by `include` (if it is
-    not ``None``).
+def _parse_excludes(column_names):
+    """Returns a pair, consisting of a list of column names to exclude on the
+    left and a dictionary mapping relation name to a list containing the names
+    of fields on the related model which should be excluded.
 
-    If `include` is ``None``, all keys will be included. If it is an iterable
-    of strings, only those keys will be included in the returned dictionary.
+    `column_names` is either ``None`` or a list of strings. If it is ``None``,
+    the returned pair will be ``(None, None)``.
+
+    If the name of a relation appears in the list then it will not appear in
+    the dictionary.
 
     """
-    result = _to_dict(instance, deep, exclude)
-    if include is None:
-        return result
-    return _include_keys(result, include)
+    if column_names is None:
+        return None, None
+    dotted_names, columns = partition(column_names, lambda name: '.' in name)
+    # Create a dictionary mapping relation names to fields on the related
+    # model.
+    relations = defaultdict(list)
+    for name in dotted_names:
+        relation, field = name.split('.', 1)
+        # Only add the relation if it's column has not been specified.
+        if relation not in columns:
+            relations[relation].append(field)
+    # Relations which are to be excluded entirely need only be in the columns
+    # list, not the relations dictionary.
+    for column in columns:
+        if column in relations:
+            del relations[column]
+    return columns, relations
 
 
 def _evaluate_functions(session, model, functions):
@@ -351,7 +437,7 @@ class FunctionAPI(ModelView):
 
         """
         try:
-            data = json.loads(request.data)
+            data = json.loads(request.args.get('q')) or {}
         except (TypeError, ValueError, OverflowError):
             return jsonify_status_code(400, message='Unable to decode data')
         try:
@@ -377,9 +463,10 @@ class API(ModelView):
     """
 
     def __init__(self, session, model, authentication_required_for=None,
-                 authentication_function=None, include_columns=None,
-                 validation_exceptions=None, results_per_page=10,
-                 post_form_preprocessor=None, *args, **kw):
+                 authentication_function=None, exclude_columns=None,
+                 include_columns=None, validation_exceptions=None,
+                 results_per_page=10, post_form_preprocessor=None, *args,
+                 **kw):
         """Instantiates this view with the specified attributes.
 
         `session` is the SQLAlchemy session in which all database transactions
@@ -409,11 +496,25 @@ class API(ModelView):
         message specifying the validation error which occurred. For more
         information, see :ref:`validation`.
 
-        `include_columns` is a list of strings which name the columns of
-        `model` which will be included in the JSON representation of that model
-        provided in response to :http:method:`get` requests. Only the named
-        columns will be included. If this list includes a string which does not
-        name a column in `model`, it will be ignored.
+        If either `include_columns` or `exclude_columns` is not ``None``,
+        exactly one of them must be specified. If both are not ``None``, then
+        the behavior of this function is undefined. `exclude_columns` must be
+        an iterable of strings specifying the columns of `model` which will
+        *not* be present in the JSON representation of the model provided in
+        response to :http:method:`get` requests.  Similarly, `include_columns`
+        specifies the *only* columns which will be present in the returned
+        dictionary. In other words, `exclude_columns` is a blacklist and
+        `include_columns` is a whitelist; you can only use one of them per API
+        endpoint. If either `include_columns` or `exclude_columns` contains a
+        string which does not name a column in `model`, it will be ignored.
+
+        If `include_columns` is an iterable of length zero (like the empty
+        tuple or the empty list), then the returned dictionary will be
+        empty. If `include_columns` is ``None``, then the returned dictionary
+        will include all columns not excluded by `exclude_columns`.
+
+        See :ref:`includes` for information on specifying included or excluded
+        columns on fields of related models.
 
         `results_per_page` is a positive integer which represents the number of
         results which are returned per page. If this is anything except a
@@ -427,6 +528,9 @@ class API(ModelView):
         is not read from the post parameters (where malicious user can tamper
         with them) but from the session.
 
+        .. versionadded:: 0.7
+           Added the `exclude_columns` keyword argument.
+
         .. versionadded:: 0.6
            Added the `results_per_page` keyword argument.
 
@@ -445,7 +549,10 @@ class API(ModelView):
         # convert HTTP method names to uppercase
         self.authentication_required_for = \
             frozenset([m.upper() for m in self.authentication_required_for])
-        self.include_columns = include_columns
+        self.exclude_columns, self.exclude_relations = \
+            _parse_excludes(exclude_columns)
+        self.include_columns, self.include_relations = \
+            _parse_includes(include_columns)
         self.validation_exceptions = tuple(validation_exceptions or ())
         self.results_per_page = results_per_page
         self.paginate = (isinstance(self.results_per_page, int)
@@ -501,7 +608,9 @@ class API(ModelView):
         of an existing model to remove. If a dictionary contains the key
         ``'id'``, that instance of the related model will be
         removed. Otherwise, the instance to remove will be retrieved using the
-        other attributes specified in the dictionary.
+        other attributes specified in the dictionary. If multiple instances
+        match the specified attributes, only the first instance will be
+        removed.
 
         If one of the dictionaries contains a mapping from ``'__delete__'`` to
         ``True``, then the removed object will be deleted after being removed
@@ -515,7 +624,6 @@ class API(ModelView):
                 subinst = self._get_by(dictionary['id'], submodel)
             else:
                 kw = unicode_keys_to_strings(dictionary)
-                # TODO document that we use .first() here
                 subinst = self.query(submodel).filter_by(**kw).first()
             for instance in query:
                 getattr(instance, relationname).remove(subinst)
@@ -629,7 +737,7 @@ class API(ModelView):
         """
         result = {}
         for fieldname, value in dictionary.iteritems():
-            if _is_date_field(self.model, fieldname):
+            if _is_date_field(self.model, fieldname) and value is not None:
                 result[fieldname] = parse_datetime(value)
             else:
                 result[fieldname] = value
@@ -719,15 +827,24 @@ class API(ModelView):
                                        message='Unable to construct query')
 
         # create a placeholder for the relations of the returned models
-        relations = _get_relations(self.model)
+        relations = frozenset(_get_relations(self.model))
+        # do not follow relations that will not be included in the response
+        if self.include_columns is not None:
+            cols = frozenset(self.include_columns)
+            rels = frozenset(self.include_relations)
+            relations &= (cols | rels)
+        elif self.exclude_columns is not None:
+            relations -= frozenset(self.exclude_columns)
         deep = dict((r, {}) for r in relations)
 
         # for security purposes, don't transmit list as top-level JSON
         if isinstance(result, list):
             return self._paginated(result, deep)
         else:
-            result = _to_dict_include(result, deep,
-                                      include=self.include_columns)
+            result = _to_dict(result, deep, exclude=self.exclude_columns,
+                              exclude_relations=self.exclude_relations,
+                              include=self.include_columns,
+                              include_relations=self.include_relations)
             return jsonify(result)
 
     # TODO it is ugly to have `deep` as an arg here; can we remove it?
@@ -739,7 +856,7 @@ class API(ModelView):
 
         `deep` is the dictionary which defines the depth of submodels to output
         in the JSON format of the model instances in `instances`; it is passed
-        directly to :func:`_to_dict_include`.
+        directly to :func:`_to_dict`.
 
         The response data is JSON of the form:
 
@@ -747,22 +864,29 @@ class API(ModelView):
 
            {
              "page": 2,
+             "total_pages": 3,
              "objects": [{"id": 1, "name": "Jeffrey", "age": 24}, ...]
            }
 
         """
+        num_results = len(instances)
         if self.paginate:
             # get the page number (first page is page 1)
             page_num = int(request.args.get('page', 1))
             start = (page_num - 1) * self.results_per_page
-            end = min(len(instances), start + self.results_per_page)
+            end = min(num_results, start + self.results_per_page)
+            total_pages = int(math.ceil(num_results / self.results_per_page))
         else:
             page_num = 1
             start = 0
-            end = len(instances)
-        objects = [_to_dict_include(x, deep, include=self.include_columns)
+            end = num_results
+            total_pages = 1
+        objects = [_to_dict(x, deep, exclude=self.exclude_columns,
+                            exclude_relations=self.exclude_relations,
+                            include=self.include_columns,
+                            include_relations=self.include_relations)
                    for x in instances[start:end]]
-        return jsonify(page=page_num, objects=objects)
+        return jsonify(page=page_num, objects=objects, total_pages=total_pages)
 
     def _check_authentication(self):
         """If the specified HTTP method requires authentication (see the
@@ -816,9 +940,20 @@ class API(ModelView):
         inst = self._get_by(instid)
         if inst is None:
             abort(404)
-        relations = _get_relations(self.model)
+        # create a placeholder for the relations of the returned models
+        relations = frozenset(_get_relations(self.model))
+        # do not follow relations that will not be included in the response
+        if self.include_columns is not None:
+            cols = frozenset(self.include_columns)
+            rels = frozenset(self.include_relations)
+            relations &= (cols | rels)
+        elif self.exclude_columns is not None:
+            relations -= frozenset(self.exclude_columns)
         deep = dict((r, {}) for r in relations)
-        result = _to_dict_include(inst, deep, include=self.include_columns)
+        result = _to_dict(inst, deep, exclude=self.exclude_columns,
+                          exclude_relations=self.exclude_relations,
+                          include=self.include_columns,
+                          include_relations=self.include_relations)
         return jsonify(result)
 
     def delete(self, instid):
@@ -890,10 +1025,19 @@ class API(ModelView):
             # Handling relations, a single level is allowed
             for col in set(relations).intersection(paramkeys):
                 submodel = cols[col].property.mapper.class_
-                for subparams in params[col]:
-                    kw = unicode_keys_to_strings(subparams)
+
+                if type(params[col]) == list:
+                    # model has several related objects
+                    for subparams in params[col]:
+                        kw = unicode_keys_to_strings(subparams)
+                        subinst = _get_or_create(self.session, submodel,
+                                                 **kw)[0]
+                        getattr(instance, col).append(subinst)
+                else:
+                    # model has single related object
+                    kw = unicode_keys_to_strings(params[col])
                     subinst = _get_or_create(self.session, submodel, **kw)[0]
-                    getattr(instance, col).append(subinst)
+                    setattr(instance, col, subinst)
 
             # add the created model to the session
             self.session.add(instance)
